﻿#include "stdafx.h"
#include <wrl\implements.h>
#include <wrl\module.h>
#include <wrl\event.h>
#include <wrl.h>
#include <Pathcch.h>
#include <windows.foundation.h>
#include <windows.foundation.numerics.h>
#include <windows.foundation.collections.h>
#include <windows.ui.input.h>
#include <wrl\client.h>
#include <winstring.h>

#include "listener.h"

#include "resource.h"

/*
	以下のヘッダファイルはcppwinrtライブラリのもの
	ここでコンパイルエラーが出る場合はインクルードディレクトリに
	ppwinrtのフォルダ（10.0.15063.0）を追加すること
*/
#include <winrt/Windows.UI.Input.h>
#include <winrt/Windows.Foundation.Collections.h>

// エラーチェックの簡略化
#define RETURN_IF_FAILED(hr) { HRESULT _hr; if(FAILED(_hr = (hr))) return _hr; }

// PNGデータのメモリバッファからRandomAccessStreamReferenceのインスタンスを作成
static HRESULT CreateRandomAccessStreamReferencFromBuffer(UINT32 bufSize, BYTE* bufptr, Microsoft::WRL::ComPtr<ABI::Windows::Storage::Streams::IRandomAccessStreamReference>& abiStreamRef)
{
	// RandomAccessStreamインスタンス作成
	Microsoft::WRL::ComPtr<ABI::Windows::Storage::Streams::IRandomAccessStream> abiIconStream;
	RETURN_IF_FAILED(Windows::Foundation::ActivateInstance(
		Microsoft::WRL::Wrappers::HStringReference(RuntimeClass_Windows_Storage_Streams_InMemoryRandomAccessStream).Get(),
		&abiIconStream));

	// アウトプット用ストリームを取得
	Microsoft::WRL::ComPtr<ABI::Windows::Storage::Streams::IOutputStream> abiOutputStream;
	RETURN_IF_FAILED(abiIconStream->GetOutputStreamAt(0, &abiOutputStream));

	// DataWriterのファクトリーを作成
	Microsoft::WRL::ComPtr<ABI::Windows::Storage::Streams::IDataWriterFactory> abiDataWriterFactory;
	RETURN_IF_FAILED(Windows::Foundation::GetActivationFactory(Microsoft::WRL::Wrappers::HStringReference(RuntimeClass_Windows_Storage_Streams_DataWriter).Get(),
		&abiDataWriterFactory));

	// DataWriter本体を作成
	Microsoft::WRL::ComPtr<ABI::Windows::Storage::Streams::IDataWriter> abiDataWriter;
	RETURN_IF_FAILED(abiDataWriterFactory->CreateDataWriter(abiOutputStream.Get(), &abiDataWriter));

	// バッファ出力
	RETURN_IF_FAILED(abiDataWriter->WriteBytes(bufSize, bufptr));

	// 出力をフラッシュする準備
	ABI::Windows::Foundation::__FIAsyncOperation_1_UINT32_t* storeOperation;
	RETURN_IF_FAILED(abiDataWriter->StoreAsync(&storeOperation));

	// 出力処理完了まで待機
	Microsoft::WRL::Wrappers::Event storeOperationCompleted(::CreateEvent(nullptr, true, false, nullptr));
	RETURN_IF_FAILED(storeOperation->put_Completed(Microsoft::WRL::Callback<ABI::Windows::Foundation::IAsyncOperationCompletedHandler<UINT32>>
		([&](ABI::Windows::Foundation::IAsyncOperation<UINT32> *pHandler, AsyncStatus status)->HRESULT
		{
			::SetEvent(storeOperationCompleted.Get());
			return S_OK;
		}).Get()));
	::WaitForSingleObject(storeOperationCompleted.Get(), INFINITE);

	// RandomAccessStreamReferenceの静的メソッド集を取得
	Microsoft::WRL::ComPtr<ABI::Windows::Storage::Streams::IRandomAccessStreamReferenceStatics> abiRandomAccessStreamReferenceStatics;
	RETURN_IF_FAILED(Windows::Foundation::GetActivationFactory(
		Microsoft::WRL::Wrappers::HStringReference(RuntimeClass_Windows_Storage_Streams_RandomAccessStreamReference).Get(),
		&abiRandomAccessStreamReferenceStatics));

	// RandomAccessStreamReferenceを作成
	Microsoft::WRL::ComPtr<ABI::Windows::Storage::Streams::IRandomAccessStreamReference> abiRandomAccessStreamReference;
	RETURN_IF_FAILED(abiRandomAccessStreamReferenceStatics->CreateFromStream(abiIconStream.Get(), &abiRandomAccessStreamReference));

	// 完了
	abiStreamRef = std::move(abiRandomAccessStreamReference);
	return S_OK;
}

// コンストラクタ
MyRadialListner::MyRadialListner()
{
	_console = GetStdHandle(STD_OUTPUT_HANDLE);

	GetConsoleScreenBufferInfo(_console, &_screenBuf);
}

// 回転検知角度を取得
HRESULT MyRadialListner::GetRotationResolutionInDegrees(double& res)
{
	return _controller->get_RotationResolutionInDegrees(&res);
}

// 回転検知角度を設定
HRESULT MyRadialListner::SetRotationResolutionInDegrees(double res)
{
	return _controller->put_RotationResolutionInDegrees(res);
}

// 回転検知時のバイブレーションのオンオフを取得
HRESULT MyRadialListner::GetUseAutomaticHapticFeedback(bool& use)
{
	boolean _use;
	HRESULT hr = _controller->get_UseAutomaticHapticFeedback(&_use);
	if(SUCCEEDED(hr))
	{
		use = (_use != 0);
	}
	return hr;
}

// 回転検知時のバイブレーションのオンオフを設定
HRESULT MyRadialListner::SetUseAutomaticHapticFeedback(bool use)
{
	return _controller->put_UseAutomaticHapticFeedback(use);
}

// 初期化処理
HRESULT MyRadialListner::Init(HWND hwnd)
{
	m_hwnd = hwnd;

	// ウィンドウに対応するRadialControllerオブジェクトを取得
	Microsoft::WRL::ComPtr<IRadialControllerInterop> controllerInterop;
	RETURN_IF_FAILED(Windows::Foundation::GetActivationFactory(
		Microsoft::WRL::Wrappers::HStringReference(RuntimeClass_Windows_UI_Input_RadialController).Get(),
		&controllerInterop));
	RETURN_IF_FAILED(controllerInterop->CreateForWindow(hwnd, IID_PPV_ARGS(&_controller)));

	// RadialControllerConfigurationを取得
	// 簡略化のためcppwinrtを使っている
	Microsoft::WRL::ComPtr<IRadialControllerConfigurationInterop> controllerConfigurationInterop;
	RETURN_IF_FAILED(Windows::Foundation::GetActivationFactory(
		Microsoft::WRL::Wrappers::HStringReference(RuntimeClass_Windows_UI_Input_RadialControllerConfiguration).Get(),
		&controllerConfigurationInterop));
	winrt::Windows::UI::Input::RadialControllerConfiguration config(nullptr); // ※cppwinrtのオブジェクト
	RETURN_IF_FAILED(controllerConfigurationInterop->GetForWindow(m_hwnd,
		IID_PPV_ARGS(winrt::put_abi(config)))); // winrt::put_abiを使うことでSDKベースとの相互運用が可能

	// 各イベントハンドラのコールバック関数を設定
	RETURN_IF_FAILED(_controller->add_ScreenContactContinued(
		Microsoft::WRL::Callback<ABI::Windows::Foundation::ITypedEventHandler<ABI::Windows::UI::Input::RadialController*, ABI::Windows::UI::Input::RadialControllerScreenContactContinuedEventArgs*>>(this, &MyRadialListner::OnScreenContactContinued).Get(),
			&_screenContactContinuedToken));
	RETURN_IF_FAILED(_controller->add_ScreenContactStarted(
		Microsoft::WRL::Callback<ABI::Windows::Foundation::ITypedEventHandler<ABI::Windows::UI::Input::RadialController*, ABI::Windows::UI::Input::RadialControllerScreenContactStartedEventArgs*>>(this, &MyRadialListner::OnScreenContactStarted).Get(),
			&_screenContactStartedToken));
	RETURN_IF_FAILED(_controller->add_ScreenContactEnded(
		Microsoft::WRL::Callback<ABI::Windows::Foundation::ITypedEventHandler<ABI::Windows::UI::Input::RadialController*,
			IInspectable*>>(this, &MyRadialListner::OnScreenContactEnded).Get(),
			&_screenContactEndedToken));
	RETURN_IF_FAILED(_controller->add_ControlLost(
		Microsoft::WRL::Callback<ABI::Windows::Foundation::ITypedEventHandler<ABI::Windows::UI::Input::RadialController*,
			IInspectable*>>(this, &MyRadialListner::OnControlLost).Get(),
			&_controlLostToken));
	RETURN_IF_FAILED(_controller->add_ControlAcquired(
		Microsoft::WRL::Callback<ABI::Windows::Foundation::ITypedEventHandler<ABI::Windows::UI::Input::RadialController*,
			ABI::Windows::UI::Input::RadialControllerControlAcquiredEventArgs*>>(this, &MyRadialListner::OnControlAcquired).Get(),
			&_controlAcquiredToken));
	RETURN_IF_FAILED(_controller->add_RotationChanged(
		Microsoft::WRL::Callback<ABI::Windows::Foundation::ITypedEventHandler<ABI::Windows::UI::Input::RadialController*,
			ABI::Windows::UI::Input::RadialControllerRotationChangedEventArgs*>>(this, &MyRadialListner::OnRotationChanged).Get(),
			&_rotatedToken));

	// ラムダ式でコールバックの内容を記述したい場合
	RETURN_IF_FAILED(_controller->add_ButtonClicked(
		Microsoft::WRL::Callback<ABI::Windows::Foundation::ITypedEventHandler<ABI::Windows::UI::Input::RadialController*,
			ABI::Windows::UI::Input::RadialControllerButtonClickedEventArgs*>>([this](ABI::Windows::UI::Input::IRadialController*,
			ABI::Windows::UI::Input::IRadialControllerButtonClickedEventArgs* args)
		{
			PrintMsg(L"ButtonClicked\n", FOREGROUND_BLUE | FOREGROUND_GREEN);
			RETURN_IF_FAILED(LogContact(args));

			return S_OK;
		}
		).Get(),
		&_buttonClickedToken));

#if 1
	// メニュー項目追加
	PopulateMenuItems();
#else
	// 単機能占有でいい場合
	winrt::Windows::UI::Input::RadialController controller(nullptr); // ※cppwinrtのオブジェクト
	_controller.CopyTo(winrt::put_abi(controller));
	config.ActiveControllerWhenMenuIsSuppressed(controller);
	config.IsMenuSuppressed(true);
#endif

	// 回転取得角度
	double res = 0;
	RETURN_IF_FAILED(GetRotationResolutionInDegrees(res));
	res = 10.0; // これを変更すると回転検知角度を変更できる
	RETURN_IF_FAILED(SetRotationResolutionInDegrees(res));

	// 回転バイブレーションを無効にする（1度単位など滑らかな入力を行う場合）
	//SetUseAutomaticHapticFeedback(false);

	// 標準メニュー項目を全削除
	winrt::Windows::Foundation::Collections::IVectorView<
		winrt::Windows::UI::Input::RadialControllerSystemMenuItemKind> items; // ※cppwinrtのオブジェクト
	config.SetDefaultMenuItems(items);

	return S_OK;
}

// メニュー項目一覧を生成
HRESULT MyRadialListner::PopulateMenuItems()
{
	RETURN_IF_FAILED(_controller->get_Menu(&_menu));
	RETURN_IF_FAILED(Windows::Foundation::GetActivationFactory(
		Microsoft::WRL::Wrappers::HStringReference(RuntimeClass_Windows_UI_Input_RadialControllerMenuItem).Get(),
		&_menuItemStatics));

	// 過去に設定されていたアイテムをクリア（特に呼ばなくても動く？）
	Microsoft::WRL::ComPtr<ABI::Windows::Foundation::Collections::IVector<ABI::Windows::UI::Input::RadialControllerMenuItem*>> menuItems;
	RETURN_IF_FAILED(_menu->get_Items(&menuItems));
	RETURN_IF_FAILED(menuItems->Clear());

	// 既知アイコンからメニュー項目追加
	RETURN_IF_FAILED(AddMenuItemFromKnownIcon(Microsoft::WRL::Wrappers::HStringReference(L"色の変更").Get(),
		ABI::Windows::UI::Input::RadialControllerMenuKnownIcon::RadialControllerMenuKnownIcon_InkColor,
		_knownIconItem1Token));
	RETURN_IF_FAILED(AddMenuItemFromKnownIcon(Microsoft::WRL::Wrappers::HStringReference(L"線の太さ").Get(),
		ABI::Windows::UI::Input::RadialControllerMenuKnownIcon::RadialControllerMenuKnownIcon_InkThickness,
		_knownIconItem2Token));

	// 絵文字グリフからメニュー項目追加
	RETURN_IF_FAILED(AddMenuItemFromSystemFont(Microsoft::WRL::Wrappers::HStringReference(L"ハート").Get(),
		Microsoft::WRL::Wrappers::HStringReference(L"Segoe UI Emoji").Get(),
		Microsoft::WRL::Wrappers::HStringReference(L"\x2764").Get(),
		_systemFontItemToken));

	// 独自アイコンからメニュー項目追加
	HINSTANCE hInst = GetModuleHandle(NULL);
	HRSRC hResPNG = FindResource(hInst, MAKEINTRESOURCE(IDB_PNG1), L"PNG");
	UINT cbResIcon = SizeofResource(hInst, hResPNG);
	BYTE* lpResIcon = (BYTE*)LockResource(LoadResource(hInst, hResPNG));
	Microsoft::WRL::ComPtr<ABI::Windows::Storage::Streams::IRandomAccessStreamReference> streamRef;
	RETURN_IF_FAILED(CreateRandomAccessStreamReferencFromBuffer(cbResIcon, lpResIcon, streamRef));
	RETURN_IF_FAILED(AddMenuItemFromIcon(Microsoft::WRL::Wrappers::HStringReference(L"独自アイコン").Get(),
		streamRef.Get(), _knownIconItem3Token));

	return S_OK;
}

// メニューアイテムの一覧への追加処理
HRESULT MyRadialListner::AddMenuItem(_In_ Microsoft::WRL::ComPtr<ABI::Windows::UI::Input::IRadialControllerMenuItem> item, _In_ HSTRING itemName, _In_ EventRegistrationToken registrationToken)
{
	// Set Microsoft::WRL::Callback
	RETURN_IF_FAILED(item->add_Invoked(
		Microsoft::WRL::Callback<ABI::Windows::Foundation::ITypedEventHandler<ABI::Windows::UI::Input::RadialControllerMenuItem*,
		IInspectable*>>(this, &MyRadialListner::OnItemInvoked).Get(),
		&registrationToken));

	// Get menu items
	Microsoft::WRL::ComPtr<ABI::Windows::Foundation::Collections::IVector<ABI::Windows::UI::Input::RadialControllerMenuItem*>> menuItems;
	RETURN_IF_FAILED(_menu->get_Items(&menuItems));

	// Add item to menu
	RETURN_IF_FAILED(menuItems->Append(item.Get()));

	// Log new item
	wchar_t message[2000];
	swprintf_s(message, 2000, L"Added %s to menu\n",
		WindowsGetStringRawBuffer(itemName, nullptr));
	PrintMsg(message, FOREGROUND_BLUE | FOREGROUND_GREEN);

	return S_OK;
}

// 独自のアイコン画像からメニュー項目を生成
HRESULT MyRadialListner::AddMenuItemFromIcon(_In_ HSTRING itemName, _In_ ABI::Windows::Storage::Streams::IRandomAccessStreamReference *icon, _In_ EventRegistrationToken registrationToken)
{
	// Create item from icon
	Microsoft::WRL::ComPtr<ABI::Windows::UI::Input::IRadialControllerMenuItem> menuItem;
	RETURN_IF_FAILED(_menuItemStatics->CreateFromIcon(itemName, icon, &menuItem));

	RETURN_IF_FAILED(AddMenuItem(menuItem, itemName, registrationToken));

	return S_OK;
}

// システム組み込みの画像からメニュー項目を生成
HRESULT MyRadialListner::AddMenuItemFromKnownIcon(_In_ HSTRING itemName, _In_ ABI::Windows::UI::Input::RadialControllerMenuKnownIcon icon, _In_ EventRegistrationToken registrationToken)
{
	// Create item from known icon
	Microsoft::WRL::ComPtr<ABI::Windows::UI::Input::IRadialControllerMenuItem> menuItem;
	RETURN_IF_FAILED(_menuItemStatics->CreateFromKnownIcon(itemName, icon, &menuItem));

	RETURN_IF_FAILED(AddMenuItem(menuItem, itemName, registrationToken));

	return S_OK;
}

// システムフォントのグリフからメニュー項目を生成
HRESULT MyRadialListner::AddMenuItemFromSystemFont(_In_ HSTRING itemName, _In_ HSTRING fontName, _In_ HSTRING itemGlyph, _In_ EventRegistrationToken registrationToken)
{
	Microsoft::WRL::ComPtr<ABI::Windows::UI::Input::IRadialControllerMenuItemStatics2> menuItemStatics2;
	RETURN_IF_FAILED(Windows::Foundation::GetActivationFactory(
		Microsoft::WRL::Wrappers::HStringReference(RuntimeClass_Windows_UI_Input_RadialControllerMenuItem).Get(),
		&menuItemStatics2));

	Microsoft::WRL::ComPtr<ABI::Windows::UI::Input::IRadialControllerMenuItem> systemFontItem;
	RETURN_IF_FAILED(menuItemStatics2->CreateFromFontGlyph(itemName, itemGlyph,
		fontName, &systemFontItem));

	RETURN_IF_FAILED(AddMenuItem(systemFontItem, itemName, registrationToken));
	return S_OK;
}

// メニューアイテムが選択された場合に呼ばれるイベント
HRESULT MyRadialListner::OnItemInvoked(_In_ ABI::Windows::UI::Input::IRadialControllerMenuItem* item, _In_ IInspectable* args)
{
	if(item != nullptr)
	{
		wchar_t message[2000];
		HSTRING str;

		item->get_DisplayText(&str);
		swprintf_s(message, 2000, L"ItemInvoked %s\n", WindowsGetStringRawBuffer(str, nullptr));
		PrintMsg(message, FOREGROUND_RED | FOREGROUND_GREEN);
	}

	return S_OK;
}

// 画面への吸着を検知（Surfaceシリーズ専用）
HRESULT MyRadialListner::OnScreenContactStarted(_In_ ABI::Windows::UI::Input::IRadialController* /*sender*/, _In_ ABI::Windows::UI::Input::IRadialControllerScreenContactStartedEventArgs* args)
{
	RETURN_IF_FAILED(LogContact(args));

	return S_OK;
}

// 画面への吸着を継続（Surfaceシリーズ専用）
HRESULT MyRadialListner::OnScreenContactContinued(_In_ ABI::Windows::UI::Input::IRadialController* /*sender*/, _In_ ABI::Windows::UI::Input::IRadialControllerScreenContactContinuedEventArgs* args)
{
	RETURN_IF_FAILED(LogContact(args));

	return S_OK;
}

// 画面から離された（Surfaceシリーズ専用）
HRESULT MyRadialListner::OnScreenContactEnded(_In_ ABI::Windows::UI::Input::IRadialController* /*sender*/, _In_ IInspectable* args)
{
	PrintMsg(L"ScreenContactEnded\n", FOREGROUND_BLUE | FOREGROUND_GREEN);

	return S_OK;
}

// ダイヤル操作を受付停止（ウィンドウがアクティブではなくなった）
HRESULT MyRadialListner::OnControlLost(_In_ ABI::Windows::UI::Input::IRadialController* /*sender*/, _In_ IInspectable* args)
{
	PrintMsg(L"ControlLost\n", FOREGROUND_RED);

	return S_OK;
}

// ダイヤル操作を受付開始（ウィンドウがアクティブになった）
HRESULT MyRadialListner::OnControlAcquired(_In_ ABI::Windows::UI::Input::IRadialController* /*sender*/, _In_ ABI::Windows::UI::Input::IRadialControllerControlAcquiredEventArgs* args)
{
	PrintMsg(L"ControlAcquired\n", FOREGROUND_GREEN);
	RETURN_IF_FAILED(LogContact(args));

	return S_OK;
}

// 回転を検知
HRESULT MyRadialListner::OnRotationChanged(_In_ ABI::Windows::UI::Input::IRadialController* /*sender*/, _In_ ABI::Windows::UI::Input::IRadialControllerRotationChangedEventArgs* args)
{
	wchar_t message[2000];
	double delta = 0;

	args->get_RotationDeltaInDegrees(&delta);
	swprintf_s(message, 2000, L"RotationChanged delta=%lf\n", delta);
	PrintMsg(message, FOREGROUND_GREEN | FOREGROUND_RED);

	RETURN_IF_FAILED(LogContact(args));

	return S_OK;
}

// 画面吸着の情報を出力
template<typename TContactArgs>
HRESULT MyRadialListner::LogContact(_In_ TContactArgs* args)
{
	// Get contact
	Microsoft::WRL::ComPtr<ABI::Windows::UI::Input::IRadialControllerScreenContact> contact;
	RETURN_IF_FAILED(args->get_Contact(&contact));

	if(contact != nullptr)
	{
		// Get contact info
		ABI::Windows::Foundation::Point position;
		ABI::Windows::Foundation::Rect bounds;
		RETURN_IF_FAILED(contact->get_Position(&position));
		RETURN_IF_FAILED(contact->get_Bounds(&bounds));

		// Log contact info
		wchar_t message[2000];
		swprintf_s(message, 2000, L"\t Postion X=%lf, Y=%lf & \n \t Bounds Height=%lf, Width=%lf, X=%lf, Y=%lf\n",
			position.X, position.Y,
			bounds.Height, bounds.Width, bounds.X, bounds.Y);

		PrintMsg(message, FOREGROUND_GREEN | FOREGROUND_BLUE);
	}

	return S_OK;
}

// コンソールにメッセージ出力
void MyRadialListner::PrintMsg(WCHAR *message, WORD messageColor)
{
	if(wcscmp(message, L"ButtonClicked\n") == 0)
	{
		SetConsoleTextAttribute(_console, 0);
	}
	UINT bufferSize = 2000;
	size_t textLength;
	DWORD charsWritten;
	StringCchLength(message, bufferSize, &textLength);

	// Set font to message color
	SetConsoleTextAttribute(_console, messageColor);

	// Send message
	WriteConsole(_console, message, (DWORD)textLength, &charsWritten, NULL);

	// Return font to original color
	SetConsoleTextAttribute(_console, _screenBuf.wAttributes);
}
